#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include "readpgm.h"

typedef struct
{
  int state;
  char magic2;
  unsigned int width,height,maxval;
  int header_ok;
  int comstate;
} pnm_header_t;

static void init_pnm_header(pnm_header_t *HH)
{
  memset(HH,0,sizeof(*HH));
}


/*** PNM header state machine
at any point, we might encounter a '#' sign. this will temporarily 
abandon the parsing until the newline is seen. after that, we restore
the old state and continue where we left off.
 **/

static int spc(unsigned char C)
{
  return isspace(C);
}

static int num(unsigned char C, unsigned int *target)
{
  if (C<'0' || C>'9') return 0;
  *target= *target*10 + (C-'0');
  return 1;
}

static int ifree(void *ptr, int dum)
{
  if (ptr) free(ptr);
  return 0;
}

static int ifclose(FILE *f,int dum)
{
  if (f) fclose(f);
  return 0;
}

static int parse_pnm_header
   (pnm_header_t *HH, unsigned char *buf, size_t len, 
    unsigned char **ds,unsigned char **de)
{
  int i;
  unsigned char C;
  if (HH->header_ok) return 0;
  for(i=0;i<len;i++)
  {
    C= buf[i];
    if (C=='#') { HH->comstate= HH->state; HH->state= 9; continue; }
    switch(HH->state)
    {
#define GOTO(x)  { HH->state= x; continue; }
#define ERROR    return 1;
#define SPC     if (spc(C))
#define NUM(X)  if (num(C,&HH->X))
#define LOOP    continue;
#define REST   
#define STATE(X) case X:

STATE(0)                       // initial, waiting for P
   if (C=='P') GOTO(1)     
   REST        ERROR   
STATE(1)                       // waiting for second magic char
   HH->magic2= C;
   GOTO(2)
STATE(2)                       // waiting for one space after magic
   SPC         GOTO(3)
   REST        ERROR
STATE(3)                       // waiting for width
   SPC         LOOP
   NUM(width)  GOTO(4)
   REST        ERROR
STATE(4)
   NUM(width)  LOOP            // calculating width
   SPC         GOTO(5)
   REST        ERROR
STATE(5)                       // waiting for height
   SPC         LOOP
   NUM(height) GOTO(6)
   REST        ERROR
STATE(6)                       // calculating height
   NUM(height) LOOP
   SPC         GOTO(7)
   REST        ERROR
STATE(7)                      // waiting for maxval
   SPC         LOOP
   NUM(maxval) GOTO(8)
   REST        ERROR
STATE(8)
   NUM(maxval) LOOP
   SPC         {
                 HH->header_ok= 1;
                 *ds= buf+i+1; *de= buf+len; 
                 return 0;
               }
   REST        ERROR
    case 9: if (C=='\n') HH->state= HH->comstate;  break;
    }
  }
  return 0;
#undef GOTO
#undef ERROR  
#undef SPC   
#undef NUM
#undef LOOP 
#undef REST   
#undef STATE
}

static char *mker(const char *fn, const char *msg,int dum)
{
  static char buf[2000];
  snprintf(buf,sizeof(buf),"%s: %s", fn, msg);
  buf[sizeof(buf)-1]= 0;
  return buf;
}

static char *mkser(const char *fn, const char *msg,int dum)
{
  static char buf[2000];
  snprintf(buf,sizeof(buf),"%s: %s (%s)", fn, msg,strerror(errno));
  buf[sizeof(buf)-1]= 0;
  return buf;
}


char *pgm_read_file
  (const char *fn, unsigned char **Rd, unsigned int *Rw,unsigned int *Rh)
{
  FILE *f;
  pnm_header_t HH;
  unsigned char buf[4096],*bs,*be,*D,*O;
  size_t R,siz,cp;

  *Rd= NULL; *Rw= *Rh= 0;

  f= fopen(fn, "rb");
  if (!f) return mkser(fn, "can't open file",0);

  init_pnm_header(&HH);
  while(1)
  {
    R= fread(buf, 1, sizeof(buf), f);
    if (R==0) return mkser(fn, "read error",ifclose(f,0));
    if (parse_pnm_header(&HH, buf, R, &bs,&be))
       return mker(fn,"bad header", ifclose(f,0));
    if (HH.header_ok) break;
  }
  if (HH.magic2!='5')
       return mker(fn,"bad file type", ifclose(f,0));
  if (HH.maxval!=255)
       return mker(fn,"bad maxval", ifclose(f,0));
  *Rw= HH.width;
  *Rh= HH.height;
  siz= HH.width*HH.height;
  D= malloc(siz);
  if (!D) 
     return mker(fn,"alloc error", ifclose(f,0));
  O= D;

  if (bs<be)
  {
    cp= be-bs; if (cp>siz) cp= siz;
    memcpy(O, bs, cp);
    O+= cp;
    siz-= cp;
  }

  while(siz)
  {
    R= fread(O, 1, siz, f);
    if (R==0) return mkser(fn,"read error",ifclose(f, ifree(D,0)));
    O+= R;
    siz-= R;
  }

  *Rd= D;

  fclose(f);
  return NULL;
}
